#define	SERIAL_PRINT
//#define	DEBUG

/*
RMEE02
A/D変換にMC3424を使ったE24系列基準の抵抗計（Arduino shield）
(C) 2022 JH4VAJ
*/

#include <Wire.h>					// I2C
#include <LiquidCrystal.h>

#define	TITLE			"RMEE02"
#define	VERSION		"1.0"
#define COPYRIGHT	"(C) 2022 JH4VAJ"

#define	RREF	10.0e3					/* 基準抵抗の値 */
#define	RMIN	5.0							/* 測定下限抵抗値 */
#define	RMAX	2.0e6						/* 測定上限抵抗値 */									

#define MC3424_ADRS		0x68		/* A/DコンバータのI2Cアドレス	*/

const int rs = 9, en = 8, d4 = 7, d5 = 6, d6 = 5, d7 = 4;
LiquidCrystal lcd(rs, en, d4, d5, d6, d7);

const float e24[] =	 {1.0, 1.1, 1.2, 1.3, 1.5, 1.6, 1.8,
											2.0, 2.2, 2.4, 2.7,
											3.0, 3.3, 3.6, 3.9,
											4.3, 4.7, 5.1, 5.6,
											6.2, 6.8, 7.5, 8.2, 9.1
											};


byte	adc_config_reg_ch_sel;	// コンフィグレジスタ: C1, C0	
int		adc_read_wait_time;			// A/Dコンバータ読出し待ち時間 [ms]

float	vadc_lsb;								// 最小電圧（1ビット分の電圧、V）

char	str_min_val[8], str_max_val[8];


// the setup function runs once when you press reset or power the board
void setup() {
	dtostrf(RMIN, 5, 2, str_min_val);			// 測定範囲下限を文字列化（表示用）

	if (RMAX < 1.0e3) {										// 測定版上限を文字列可（k、Mの接頭辞対応）
		dtostrf(RMAX, 6, 2, str_max_val);
	}
	else if (RMAX < 1.0e6) {
		dtostrf((RMAX / 1.0e3), 6, 2, str_max_val);
		strcat(str_max_val, "k");
	}
	else {
		dtostrf((RMAX / 1.0e6), 6, 2, str_max_val);
		strcat(str_max_val, "M");
	}

	adc_config_reg_ch_sel = 0x0C;			// A/Dコンバータビット数設定（18ビット）
	adc_read_wait_time = 300;					// A/Dコンバータ読出し待ち時間[ms]
	vadc_lsb = 2.048 / (float)((long)1 << (18 - 1));	// ADC １bitあたりの電圧（2のべき乗をシフトで表現）
			
#ifdef SERIAL_PRINT
	Serial.begin(9600);
#endif

	lcd.begin(16, 2);
	
	print_version();

	Wire.begin();

	check_voltage();
	delay(1500);		// 表示待ち時間
}


void print_version() {
#ifdef SERIAL_PRINT
	Serial.print(TITLE);
	Serial.print(" Ver.");
	Serial.println(VERSION);
	Serial.println(COPYRIGHT);
#endif

	lcd.clear();
	lcd.print(TITLE);
	lcd.print(" Ver.");
	lcd.print(VERSION);
	lcd.setCursor(0, 1);
	lcd.print(COPYRIGHT);
	delay(1000);
}



void check_voltage() {
	float v18_adc;				// MCP3424の1.8V（実測値）
	char	str_v18_adc[8];
	float	vcc_mcu;				// ATmega328PのVcc
	char	str_vcc_mcu[8];

	v18_adc = (float)read_adc(4, 1) * vadc_lsb;
	dtostrf(v18_adc, 5, 3, str_v18_adc);
	
	vcc_mcu = (v18_adc / (float)analogRead(A0)) * 1024.0;	// v18_adcを元に逆算
	dtostrf(vcc_mcu, 5, 3, str_vcc_mcu);

#ifdef SERIAL_PRINT
	Serial.print("Vcc: ");
	Serial.print(str_vcc_mcu);
	Serial.println("V");
	Serial.print("V18: ");
	Serial.print(str_v18_adc);
	Serial.println("V");
#endif

	lcd.clear();
	lcd.print("Vcc:");
	lcd.print(str_vcc_mcu);
	lcd.print("V");
	lcd.setCursor(0, 1);
	lcd.print("V18:");
	lcd.print(str_v18_adc);
	lcd.print("V");
}


/*
 * A/Dコンバータからの読出し
 * 
 * ch:	1, 2, 3, 4
 * pga: 1, 2, 4, 8		// 内蔵アンプの倍率
 */
 
long read_adc(byte ch, byte pga) {
	const byte	adrs = MC3424_ADRS;
	const byte	num_dat = 4;				// データ読出しの回数（18ビットの場合だけ4回。それ以外は3回）
	
	long	dat;
	byte	adc_config_reg;
	int		i;

	switch (ch) {
		case 4:
			adc_config_reg = adc_config_reg_ch_sel | 0x70;
			break;
		case 3:
			adc_config_reg = adc_config_reg_ch_sel | 0x50;
			break;
		case 2:
			adc_config_reg = adc_config_reg_ch_sel | 0x30;
			break;
 		case 1:
		default:
			adc_config_reg = adc_config_reg_ch_sel | 0x10;
			break;
	}

	switch (pga) {
		case 8:
			adc_config_reg |= 0x03;	// x8
			break;
		case 4:
			adc_config_reg |= 0x02;	// x4
			break;
		case 2:
			adc_config_reg |= 0x01;	// x2
			break;
		case 1:
		default:
			adc_config_reg |= 0x00;	// x1
			break;
	}
	
	Wire.beginTransmission(adrs);
	Wire.write(adc_config_reg);
	Wire.endTransmission(false);

	delay(adc_read_wait_time);

	Wire.requestFrom(adrs, num_dat);

	/* データ読み出し、および、整形 */
	dat = 0;
	for (i = 0; i < (num_dat - 1); i++) {
		dat <<= 8;
		dat |= Wire.read();
	}
	
	adc_config_reg = Wire.read();

	if ((adc_config_reg & 0x80) != 0x00) {
		print_error("ADC read error");
		
		while(1) {}	// エラー時は動作停止
	}

	if ((dat & 0x00040000) != 0) {	// if (MSB == HIGH) {dat = 0;}
		dat = 0;											// マイナスの値なら0を返す
	}

	return(dat);
}



/*
 * A/Dコンバータからの読出し（8倍値）
 *   ADCの値が小さい値なら8倍精度で読み出す
 */

long read_adc_8(byte ch) {
	long	dat;
	const long	thld_8 = 0x00000800;	// 8倍読出しを行う閾値
	
	dat = read_adc(ch, 1);		// A/Dコンバータから読み出す（1倍）

	if (dat > thld_8) {				// if (読み出した値が閾値よりも大きい)
		dat <<= 3;								// 8倍にする
	}
	else {
		dat = read_adc(ch, 8);		// 内蔵アンプを有効化して8倍精度で読み出す
	}
	
	return(dat);
}


void measure_resistance() {
	byte	pga;
	long	val_ref, val_tgt, val_diff;
	float	r;
	char	msg[16];

	/* Read reference (CH4) */
	val_ref = read_adc_8(4);

	/* Read target (CH3)*/
	val_tgt = read_adc_8(3);

	val_diff = val_ref - val_tgt;
	val_diff = (val_diff < 64) ? 0 : val_diff;	// 差が一定値より小さい場合は0と見なす

#if defined(DEBUG) && defined(SERIAL_PRINT)
	Serial.print("val_ref: ");
	Serial.print(val_ref);
	Serial.print("	val_tgt: ");
	Serial.print(val_tgt);
	Serial.print("	val_dif: ");
	Serial.println(val_diff);
#endif

	if (0 < val_diff) {
		r = ((float)val_tgt / (float)val_diff) * RREF;
	}
	else {				// val_ref < val_tgt（A/Dコンバータの誤差など）
		r = RMAX + 1.0;
	}

	if (r < RMIN) {
		sprintf(msg, "Rx < %s", str_min_val);
		print_error(msg);
	}
	else if (RMAX < r) {
		if (val_diff < 64) {
			sprintf(msg, "Open (too high)");
			print_error(msg);
		}
		else {
			sprintf(msg, "%s < Rx", str_max_val);
			print_error(msg);
		}
	}
	else {
		print_result(r);
	}
}




void print_error(char* str) {
#ifdef SERIAL_PRINT
	Serial.println(str);
#endif
	lcd.clear();
	lcd.print(str);
}



void print_result(float r) {
	char	str_buf[10];			// 表示文字列バッファ
	char	*str_kM;					// 接頭辞
	float	kM;								// 接頭辞に合わせた倍率
	float	match_e24_kM;			// マッチしたE24系列の値（接頭辞考慮）
	float	multi;						// E24系列の基数に掛け合わせる倍率
	float	diff_0, diff_1;		// E24系列マッチ判定用
	float	diff_kM;					// E24系列からの差分（接頭辞考慮）
	float	diff_cent;				// E24系列からの差分（百分率）

	const int digi_mes = 5;	// 測定結果表示桁数
	int		n_digi;						// 整数部の表示桁数

	const float	r_thsld = (e24[0] + (e24[(sizeof(e24) / sizeof(float)) - 1] / 10.0)) / 2.0;
													// 桁上げ判定閾値: (1.0 + 0.91) / 2.0）= 0.955
	
	int		i;


	if (r > (r_thsld * 1.0e6)) {	// 接頭辞（k, M）判定
		str_kM = "M";
		kM = 1.0e6;
	}
	else if (r > (r_thsld * 1.0e3)) {
		str_kM = "k";
		kM = 1.0e3;
	}
	else {
		str_kM = "";
		kM = 1.0;
	}

	if (r < (r_thsld * 1.0)) {		// 乗数と整数部の表示桁数判定
		multi = 0.1;
		n_digi = 1;
	}
	else if (r < (r_thsld * 10.0)) {
		multi = 1.0;
		n_digi = 1;
	}
	else if (r < (r_thsld * 100.0)) {
		multi = 10.0;
		n_digi = 2;
	}
	else if (r < (r_thsld * 1.0e3)) {
		multi = 100.0;
		n_digi = 3;
	}
	else if (r < (r_thsld * 10.0e3)) {
		multi = 1.0e3;
		n_digi = 1;
	}
	else if (r < (r_thsld * 100.0e3)) {
		multi = 10.0e3;
		n_digi = 2;
	}
	else if (r < (r_thsld * 1.0e6)) {
		multi = 100.0e3;
		n_digi = 3;
	}
	else if (r < (r_thsld * 10.0e6)) {
		multi = 1.0e6;
		n_digi = 1;
	}
	else if (r < (r_thsld * 100.0e6)) {
		multi = 10.0e6;
		n_digi = 2;
	}
	else {
		multi = 100.0e6;
		n_digi = 3;
	}

	/* E24系列判定 */
	diff_0 = 1.0e30;		// ダミーの最大値（初期値）

	for (i = 0; i < (sizeof(e24) / sizeof(float)); i++) {	
		diff_1 = r - (e24[i] * multi);
		if (abs(diff_0) < abs(diff_1)) {		// E24系列からの差分が最小のものを見つける
			break;
		}
		else {
			diff_0 = diff_1;
		}
	}

	match_e24_kM = e24[i - 1] * (multi / kM);

	/* 表示 */
	
	lcd.clear();

	/* 測定値 */
	dtostrf((r / kM), digi_mes, (digi_mes - 1 - n_digi), str_buf);
#ifdef SERIAL_PRINT
	Serial.print(str_buf);
	Serial.print(str_kM);
	Serial.print("Ω");
#endif
  lcd.print(str_buf);
  lcd.print(str_kM);
  lcd.write(0xf4);		// Ω

	/* E24系列からの差分 */
	diff_kM = diff_0 / kM;
	dtostrf(diff_kM, 7, 3, str_buf);
#ifdef SERIAL_PRINT
	Serial.print(" ");
	Serial.print(str_buf);
	Serial.print(str_kM);
	Serial.print("Ω");
#endif
  lcd.setCursor(9, 0);
  dtostrf(diff_kM, 7, 3, str_buf);
  lcd.print(str_buf);

	/* マッチするE24系列の値 */
	if (n_digi == 1) {				// 整数部が一桁なら小数点以下は一桁（e.g. 4.7kΩ）
		dtostrf(match_e24_kM, 3, 1, str_buf);
	}
	else {										// それ以外は小数点以下はなし（e.g. 33kΩ）
		dtostrf(match_e24_kM, n_digi, 0, str_buf);
	}
	lcd.setCursor(0, 1);
  lcd.print(str_buf);
  lcd.print(str_kM);
	lcd.print(" E24");
#ifdef SERIAL_PRINT
	Serial.print("  E24_");
	Serial.print(str_buf);
	Serial.print(str_kM);
	Serial.print("Ω");
#endif

	/* 誤差 */
	diff_cent = (diff_kM / match_e24_kM) * 100.0;
	dtostrf(diff_cent, 6, 2, str_buf);
	lcd.setCursor(9, 1);
  lcd.print(str_buf);
  lcd.print("%");
#ifdef SERIAL_PRINT
	Serial.print(" ");
	Serial.print(str_buf);
	Serial.print("%");

	Serial.println("");
#endif
}


// the loop function runs over and over again forever
void loop() {
	measure_resistance();

	delay(300);		// A/Dコンバータの読出し待ち時間もあるので短めに
}
